{===============================================================================
  Unit.................MLP.pas
  Description..........TMLP component for using classical Multilayer Perceptron
  Accessed Units.......Classes, SysUtils, Dialogs
  Compiler.............Delphi 3,4,5,6,7
  References...........Haykin, Simon. Neural Networks-A Comprehensive Foundation.
                       1994.
  Author...............Luciano F. de Medeiros, lfm@netpar.com.br
  Conditions of usage..Freeware and Open Source, use at own risk. Please report
                       faults or comments to the author, and any modifications
                       in the original source
================================================================================}
unit MLP;

interface

uses Classes, SysUtils, Dialogs, Math;

type
  TLayerType = (ltInput, ltHidden, ltOutput);

  PLayer = ^TLayer;
  PNeuron = ^TNeuron;
  PSynapsis = ^TSynapsis;

  TLayer = record
     Number : integer; // identificao da camada
     LayerType : TLayerType; // tipo de camada
     LayerNext : PLayer; // aponta para prxima camada
     LayerPrior : pointer; // aponta para camada anterior
     NeuronFirst : pointer;  // aponta para o primeiro TNeuron da camada
     SynapsisFirst : pointer; // aponta para a primeira TSynapsis da camada
     NeuronCount : integer; // quantidade de neurnios da camada
  end;

  TNeuron = record
     Number : integer; // identificao do neurnio
     Value : double; // valor assumido pelo neurnio
     Delta : double;  // valor para gradiente descendente
     Target : double; // valor de alvo utilizado somente para ltima camada
     NeuronNext : pointer; // ponteiro para proximo TNeuron da mesma camada
     MaxValue: double; // mximo valor assumido pelo neurnio
     MinValue: double; // mnimo valor assumido pelo neurnio
     Bias: double; // valor de bias
  end;

  TSynapsis = record
     Number : longint; // identificao da sinapse
     Value : double;  // valor assumido pela sinapse
     Delta : double; // refere-se ao delta de correo dos pesos sinpticos
     NeuronPrior, NeuronNext : PNeuron; // ponteiros ligando os neurnios
     SynapsisNext : PSynapsis; // ponteiro para prxima TSynapsis
  end;

  TMLP = class(TComponent)
  private
     FStruct : TStrings; // acessa a estrutura de uma lista
     FEpochs : longint; // nmero de pocas para o treinamento
     FMomentum : double; // coeficiente de momentum
     FLearning : double; // coeficiente de aprendizado
     FFlatness : double; // inflexo da funo sigmide
     FKnowledge : string; // nome do arquivo de pesos
     procedure SetStruct(Value : TStrings);
     procedure SetEpochs(Value: longint);
     procedure SetMomentum(Value: double);
     procedure SetLearning(Value: double);
     procedure SetFlatness(Value: double);
     procedure SetKnowledge(Value: string);
  protected
     FGamma : double;
     Lista : string;
     Prepared : boolean;
     Layer, LastLayer : PLayer; // primeira e ltima camada
     function GetLayer : PLayer;
     procedure FreeLayer(var Node : PLayer);
     function GetNeuron : PNeuron;
     procedure FreeNeuron(var Node : PNeuron);
     function GetSynapsis : PSynapsis;
     procedure FreeSynapsis(var Node : PSynapsis);
     procedure FeedForward; // avana na rede calculando as sadas
     function Activation(Value: double) : double; // funo sigmide
     procedure BackPropagation; // calcula deltas aps feedforward
     procedure CorrectWeight; // corrige pesos aps backpropagation
     procedure CorrectWeightPruning;
     function Eliminate : longint; // para ver a quantidade de pesos zerados
  public
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     procedure Build; // constri a rede ( preciso rodar sempre)
     procedure Training; // executa uma poca de treinamento da rede
     function TrainingPruning(Epochs: longint) : longint;
     procedure SetInput(Neuron: integer; Value: double); // marca entradas na primeira camada
     procedure SetInputMax(Neuron: integer; Value: double); // marca valor mximo na primeira camada
     procedure SetInputMin(Neuron: integer; Value: double); // marca valor mnimo na primeira camada
     procedure SetOutput(Neuron: integer; Value: double); // marca neurnios na saida
     procedure SetOutputMax(Neuron: integer; Value: double); // valor mximo nas sadas
     procedure SetOutputMin(Neuron: integer; Value: double); // valor mnimo nas sadas
     function GetOutput(Neuron: integer) : double; // pega valores de sada
     function GetWeight(LayerNumber: Integer; Value: Integer): Double;
     procedure SetBias(L, N: integer; Value: double); // ajusta valor do bias (Entrada : L = 1)
     procedure Test; // Valores para teste de valores de entrada
     procedure Clear; // zera os valores das sinapses
     procedure Save; // salva sinapses em arquivo
     procedure Load; // carrega sinapses em arquivo, com rede j construda com BUILD
     function Cost : double; // retorna a funo custo, ou seja, o erro atual
     function List : string; // mtodo provisrio, s para controle
     procedure Purge; // elimina os pesos zerados
  published
     property Struct : TStrings read FStruct write SetStruct; // estrutura da rede
     property Momentum : double read FMomentum write SetMomentum;
     property Learning : double read FLearning write SetLearning;
     property Knowledge : string read FKnowledge write SetKnowledge;
     property Flatness : double read FFlatness write SetFlatness;
  end;

procedure Register;

implementation

procedure TMLP.Build;
var P, Q : PLayer;
    R, S : PNeuron;
    T, U : PSynapsis;
    i, j : integer;
    k : longint;
    First : boolean;
begin
   if not Prepared then
   begin
      // Cria primeira camada
      Layer := GetLayer; // ponteiro para camada inicial
      Layer.Number := 1;
      Layer.LayerType := ltInput;
      Layer.LayerPrior := nil;

      // Cria neurnios da primeira camada
      R := GetNeuron;
      R.Number := 1;
      R.Value := 0;
      Layer.NeuronFirst := R;
      Layer.NeuronCount := StrToInt(FStruct[0]);
      for i := 1 to StrToInt(FStruct[0])-1 do
      begin
         S := GetNeuron;
         R.NeuronNext := S;
         S.Number := i+1;
         R := S;
      end;
      R.NeuronNext := nil;

      // Cria prximas camadas
      P := Layer;
      for i := 1 to FStruct.Count-1 do
      begin
         Q := GetLayer;
         P.LayerNext := Q;
         Q.LayerPrior := P;
         Q.Number := i+1;
         Q.NeuronCount := StrToInt(FStruct[i]);
         if (i = FStruct.Count-1) then
            Q.LayerType := ltOutput
         else
            Q.LayerType := ltHidden;
         Q.LayerNext := nil;
         P := Q;

         // Cria neurnios das camadas
         R := GetNeuron;
         R.Number := 1;
         R.Value := 0;
         R.Delta := 1;
         P.NeuronFirst := R;
         for j := 1 to StrToInt(FStruct[i])-1 do
         begin
            S := GetNeuron;
            R.NeuronNext := S;
            S.Number := j+1;
            S.Value := 0;
            S.Delta := 1;
            R := S;
         end;
         R.NeuronNext := nil;
      end;
      LastLayer := Q; // guarda ponteiro para a ltima camada

      // Cria sinapses entre neurnios. A partir daqui no  preciso mais a quantidade
      // de cada neurnio.
      Randomize;
      P := Layer;
      Q := Layer.LayerNext;
      k := 1;
      while Q <> nil do
      begin
         R := P.NeuronFirst;
         S := Q.NeuronFirst;
         First := true;
         while R <> nil do
         begin
            while S <> nil do
            begin
               T := GetSynapsis;
               if not First then
                  U.SynapsisNext := T // monta a cadeia com os seguintes
               else
                  P.SynapsisFirst := T; // liga a primeira sinapse  camada
               First := false;
               T.Number := k;
               T.Value := Random(1000)/1000 - 0.5; // comea com valores aleatrios
//               T.Value := Random(100) - 50; // comea com valores aleatrios
//               T.Value := Random(10) - 5; // comea com valores aleatrios
//               T.Value := 0;
               T.Delta := Random(1000)/1000 - 0.5; // comea com deltas iguais a zero
               T.NeuronPrior := R;
               T.NeuronNext := S;
               U := T;
               S := S.NeuronNext;
               Inc(k);
            end;
            R := R.NeuronNext;
            S := Q.NeuronFirst;
         end;
         P := Q; // Assume P como prxima camada e Q como a seguinte
         Q := Q.LayerNext;
         T.SynapsisNext := nil;
      end;
      T.SynapsisNext := nil;
      Prepared := True;
   end
   else
      ShowMessage('Rede no construda');

end;

procedure TMLP.FeedForward;
var P, Q : PLayer;
    R : PNeuron;
    T : PSynapsis;
begin
   P := Layer;
   Q := P.LayerNext;
   R := P.NeuronFirst;
   while Q <> nil do
   begin
      // Primeira sinapse da camada
      T := P.SynapsisFirst;
      // Acumula valores de w * x na prxima camada
      while T <> nil do
      begin
         T.NeuronNext.Value := T.NeuronNext.Value + T.NeuronPrior.Value * T.Value + T.NeuronPrior.Bias; // acumula v = w * x + bias;
         T := T.SynapsisNext;
      end;
      // Normaliza valores de v, calculando a ativao
      R := Q.NeuronFirst;
      while R <> nil do
      begin
         R.Value := Activation(R.Value / (P.NeuronCount));
         R := R.NeuronNext;
      end;
      // Prxima camada
      P := Q;
      Q := Q.LayerNext;
   end;
end;

procedure TMLP.BackPropagation;
var P, Q : PLayer;
    R : PNeuron;
    T : PSynapsis;
begin
   Q := LastLayer;
   R := Q.NeuronFirst;
   // Calcula deltas para camada de sada
   while R <> nil do
   begin
      R.Delta := R.Value * (1 - R.Value) * (R.Target - R.Value);
      R := R.NeuronNext;
   end;

   // Calcula deltas para camadas internas
   Q := Q.LayerPrior;
   while Q.LayerPrior <> nil do
   begin
      T := Q.SynapsisFirst;
      while T <> nil do
      begin
         T.NeuronPrior.Delta := T.NeuronPrior.Delta + T.Value * T.NeuronNext.Delta;
         T := T.SynapsisNext;
      end;
      Q := Q.LayerPrior;
   end;

   // Normaliza deltas para camadas ocultas, a de sada no  preciso
   Q := LastLayer.LayerPrior;
   while Q.LayerPrior <> nil do
   begin
      R := Q.NeuronFirst;
      while R <> nil do
      begin
         R.Delta := R.Value * (1 - R.Value) * R.Delta / Q.LayerNext.NeuronCount;
         R := R.NeuronNext;
      end;
      Q := Q.LayerPrior;
   end;
end;

procedure TMLP.CorrectWeight;
var P : PLayer;
    T : PSynapsis;
begin
   P := Layer;
   // Corrige os pesos de cada camada
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         T.Delta := (FMomentum * T.Delta) - (FLearning * T.NeuronNext.Delta * T.NeuronPrior.Value);
         T.Value := T.Value - T.Delta;  // atualiza as sinapses
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
end;

procedure TMLP.CorrectWeightPruning;
var P : PLayer;
    T : PSynapsis;
    Square : double;
begin
   P := Layer;
   // Corrige os pesos de cada camada
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         Square := T.Value * T.Value;
         T.Value := T.Value * (1 -  (FGamma / ((1+Square)*(1+Square))));
         T.Delta := (0.9 * T.Delta) - (T.NeuronNext.Delta * T.NeuronPrior.Value);
         T.Value := T.Value - T.Delta;  // atualiza as sinapses
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
end;

procedure TMLP.Test;
begin
   FeedForward; // faz somente a passagem dos valores de entrada pelos pesos sinpticos
end;

procedure TMLP.Clear;
var P : PLayer;
    T : PSynapsis;
begin
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
//         T.Value := Random(1000)/1000 - 0.5;
         T.Value := Random(10) - 5;
//         T.Value := 0;
         T.Delta := 0;
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
end;

function TMLP.GetWeight(LayerNumber: Integer; Value: Integer): Double;
var P : PLayer;
    T : PSynapsis;
label return;
begin
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         if (T.Number = Value) and (P.Number = LayerNumber) Then
         begin
            Result := T.Value;
            goto return;
         end;
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
return:
End;

function TMLP.Activation(Value: double) : double;
begin
   Result := 1/(1+exp(-Flatness * Value));
end;

function TMLP.List : string;
var P : PLayer;
    R : PNeuron;
    T : PSynapsis;
    Temp: string;
begin
   Lista := '';
   P := Layer; // pega o apontador do perceptron - 1 camada
   while P <> nil do
   begin
      Lista := Lista + 'C: '+ IntToStr(P.Number) + Chr(13) + Chr(10);
      R := P.NeuronFirst;
      while R <> nil do
      begin
         Lista := Lista + ' N: '+ IntToStr(R.Number) + Chr(13) + Chr(10);
         R := R.NeuronNext;
      end;
      P := P.LayerNext;
   end;

   Lista := Lista + ' ##### ';
   P := Layer; // pega o apontador do perceptron - 1 camada
   while P.LayerNext <> nil do
   begin
      Lista := Lista + 'C: '+ IntToStr(P.Number) + Chr(13) + Chr(10);
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         Temp := Format('%4.4f',[T.Value]);
         Lista := Lista + ' S: '+ Format('%3.3d',[T.Number]) + '/' +
         Format('%3.3d',[T.NeuronPrior.Number]) + '-' + Format('%3.3d',[T.NeuronNext.Number]) + ': ' +
         StringOfChar(' ',10-Length(Temp)) + Temp +
         Chr(13) + Chr(10);
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
   Result := Lista;
end;

function TMLP.Eliminate : longint;
var Zeros : longint;
    P : PLayer;
    T : PSynapsis;
begin
   Zeros := 0;
   P := Layer; // pega o apontador do perceptron - 1 camada
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         if Trunc(T.Value*100) = 0 then Inc(Zeros);
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
   Result := Zeros;
end;

procedure TMLP.Training;
begin
   // Presume que os valores de entrada j estejam colocados, junto com os alvos
   // Executa o processo somente uma vez
   FeedForward;
   BackPropagation;
   CorrectWeight;
end;

function TMLP.TrainingPruning(Epochs: longint) : longint;
begin
   // Presume que os valores de entrada j estejam colocados, junto com os alvos
   // Executa o processo somente uma vez
   FeedForward;
   BackPropagation;
   FGamma := FGamma + (0.1/(Epochs*2));
   CorrectWeightPruning;
   Result := Eliminate; // conta quantas sinapses foram zeradas pelo processo
end;

function TMLP.Cost : double;
var Error: double;
    P : PLayer;
    R : PNeuron;
begin
   P := LastLayer;
   R := P.NeuronFirst;
   Error := 0;
   while R <> nil do
   begin
      Error := Error + (R.Target - R.Value)*(R.Target - R.Value);
      R := R.NeuronNext;
   end;
   Result := 0.5 * Error;
end;

procedure TMLP.SetInput(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := Layer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         // entrada j transformada
         R.Value := 1 + 2 * (Value - R.MaxValue) / (R.MaxValue - R.MinValue);
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetInputMax(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := Layer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         R.MaxValue := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetInputMin(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := Layer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         R.MinValue := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetOutput(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := LastLayer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         // sada j transformada
//         R.Target := 1 + 2 * (Value - R.MaxValue) / (R.MaxValue - R.MinValue); // [-1,1]
         R.Target := 1 + (Value - R.MaxValue) / (R.MaxValue - R.MinValue); // [0,1]
//         R.Target := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetOutputMax(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := LastLayer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         R.MaxValue := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetOutputMin(Neuron: integer; Value: double);
var P : PLayer;
    R : PNeuron;
begin
   P := LastLayer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
         R.MinValue := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

function TMLP.GetOutput(Neuron: integer) : double;
var P : PLayer;
    R : PNeuron;
begin
   P := LastLayer;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = Neuron then
      begin
//         Result := R.Value;
//         Result := 0.5 * (R.Value - 1) * (R.MaxValue - R.MinValue) + R.MaxValue; // [-1,1]
         Result := (R.Value - 1) * (R.MaxValue - R.MinValue) + R.MaxValue; // [0,1]
         Result := Result + R.Bias;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.SetBias(L, N: integer; Value: double); // ajusta valor do bias
var P : PLayer;
    R : PNeuron;
    La: integer;
begin
   P := Layer;
   La := 1;
   while La <> L do
   begin
      P := P.LayerNext;
      Inc(La);
   end;
   R := P.NeuronFirst;
   while R <> nil do
   begin
      if R.Number = N then
      begin
         R.Bias := Value;
         break;
      end;
      R := R.NeuronNext;
   end;
end;

procedure TMLP.Save;
var F: TextFile;
    P : PLayer;
    T : PSynapsis;
begin
   AssignFile(F, FKnowledge);
   Rewrite(F);
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         WriteLn(F, Format('%d',[T.Number]));
         WriteLn(F, Format('%d',[T.NeuronPrior.Number]));
         WriteLn(F, Format('%d',[T.NeuronNext.Number]));
         WriteLn(F, Format('%d',[P.Number]));
         WriteLn(F, Format('%4.12f', [T.Value]));
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
   CloseFile(F);
end;

procedure TMLP.Load;
var F: TextFile;
    P, Q : PLayer;
    R, S : PNeuron;
    T, U : PSynapsis;
    SynValue, SynNumber, SynNeuronPrior, SynNeuronNext, SynLayer, SynLayerAnt : string;
    First : boolean;
label 1;
begin
   // Primeiro destri as sinapses
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         P.SynapsisFirst := T.SynapsisNext; // pula o primeiro
         FreeSynapsis(T); // libera o primeiro
         T := P.SynapsisFirst;
      end;
      P := P.LayerNext;
   end;

   // Abre arquivo netfile
   AssignFile(F, FKnowledge);
   Reset(F);

   // Depois, reconstri as conexes sinpticas
   First := true;
   while not Eof(F) do
   begin
      // L do arquivo
      ReadLn(F, SynNumber);
      ReadLn(F, SynNeuronPrior);
      ReadLn(F, SynNeuronNext);
      ReadLn(F, SynLayer);
      ReadLn(F, SynValue);

      if (StrToInt(SynLayer) > 1) and (SynLayer <> SynLayerAnt) then
      begin
         T.SynapsisNext := nil;
         First := true;
      end;
      SynLayerAnt := SynLayer;

      P := Layer;
      Q := Layer.LayerNext;
      while Q <> nil do
      begin
         R := P.NeuronFirst;
         while R <> nil do
         begin
            S := Q.NeuronFirst;
            while S <> nil do
            begin
               if (R.Number = StrToInt(SynNeuronPrior)) and (S.Number = StrToInt(SynNeuronNext)) and (P.Number = StrToInt(SynLayer)) then
               begin
                  T := GetSynapsis;
                  if not First then
                     U.SynapsisNext := T // monta a cadeia com os seguintes
                  else
                     P.SynapsisFirst := T; // liga a primeira sinapse  camada
                  First := false;

                  T.Number := StrToInt(SynNumber);
                  T.Value := StrToFloat(SynValue);
                  T.Delta := 0;
                  T.NeuronPrior := R;
                  T.NeuronNext := S;
                  U := T;
                  goto 1;
               end;
               S := S.NeuronNext;
            end;
            R := R.NeuronNext;
         end;
         P := Q; // Assume P como prxima camada e Q como a seguinte
         Q := Q.LayerNext;
      end;
1:
   end;
   T.SynapsisNext := nil;
   CloseFile(F);
end;

procedure TMLP.Purge;
var P : PLayer;
    T, U, V : PSynapsis;
    k : longint;
    Factor: integer;
begin
   // Expurga os pesos zerados
   Factor := 100;
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      U := T;
      T := T.SynapsisNext;
      while T <> nil do
      begin
         if Trunc(T.Value*Factor) = 0 then
         begin
            U.SynapsisNext := T.SynapsisNext;
            V := T;
            T := T.SynapsisNext;
            FreeSynapsis(V);
         end
         else
         begin
            U := T;
            T := T.SynapsisNext;
         end;
      end;
      P := P.LayerNext;
   end;

   // Renumera os pesos
   k := 1;
   P := Layer;
   while P.LayerNext <> nil do
   begin
      T := P.SynapsisFirst;
      while T <> nil do
      begin
         T.Number := k;
         Inc(k);
         T := T.SynapsisNext;
      end;
      P := P.LayerNext;
   end;
end;

function TMLP.GetLayer : PLayer;
var P : PLayer;
begin
   New(P);
   Result := P;
end;

procedure TMLP.FreeLayer(var Node : PLayer);
begin
   Dispose(Node);
end;

function TMLP.GetNeuron : PNeuron;
var P : PNeuron;
begin
   New(P);
   P.Bias := 0;
   Result := P;
end;

procedure TMLP.FreeNeuron(var Node : PNeuron);
begin
   Dispose(Node);
end;

function TMLP.GetSynapsis : PSynapsis;
var P : PSynapsis;
begin
   New(P);
   Result := P;
end;

procedure TMLP.FreeSynapsis(var Node : PSynapsis);
begin
   Dispose(Node);
end;

procedure TMLP.SetStruct(Value: TStrings);
begin
   if Value.Count < 2 then
      ShowMessage('Uma rede MLP deve ter no mnimo 2 camadas')
   else
      FStruct.Assign(Value);
end;

procedure TMLP.SetEpochs(Value: longint);
begin
   FEpochs := Value;
end;

procedure TMLP.SetMomentum(Value: double);
begin
   FMomentum := Value;
end;

procedure TMLP.SetLearning(Value: double);
begin
   FLearning := Value;
end;

procedure TMLP.SetKnowledge(Value: string);
begin
   FKnowledge := Value;
end;

procedure TMLP.SetFlatness(Value: double);
begin
   FFlatness := Value;
end;

constructor TMLP.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FStruct := TStringList.Create;
   SetMomentum(0.5);
   SetLearning(0.9);
   SetKnowledge('');
   SetFlatness(1);
   Prepared := false;
   FGamma := 0;
end;

destructor TMLP.Destroy;
begin
   FStruct.Free;
   inherited Destroy;
end;

procedure Register;
begin
   RegisterComponents('Lfm', [TMLP]);
end;

end.
